from pathlib import Path
import sys
from typing import Union


class HttpOnlyCookie_Handler:
    '''
    如果 python<=3.10.6 ，去除 HTTP-Only cookie
    bpo-38976
    '''
    cookies_path: Union[str, Path]
    cookies_raw_str: str

    def __init__(self, cookies_path: Union[str, Path]):
        self.cookies_path = cookies_path
    def __enter__(self):
        if sys.version_info < (3, 10, 6):
            print('python<=3.9，去除 HTTP-Only cookie')
            HTTPONLY_PREFIX = "#HttpOnly_"
            with open(self.cookies_path, 'r', encoding='utf-8') as f:
                self.cookies_raw_str = f.read()
            cookies_new_str = self.cookies_raw_str.replace(HTTPONLY_PREFIX, '')
            with open(self.cookies_path, 'w', encoding='utf-8') as f:
                f.write(cookies_new_str)
    def __exit__(self, exc_type, exc_value, traceback):
        if sys.version_info < (3, 10, 6):
            print('python<=3.9，恢复 HTTP-Only cookie')
            with open(self.cookies_path, 'w', encoding='utf-8') as f:
                f.write(self.cookies_raw_str)



import asyncio
from contextlib import asynccontextmanager
from datetime import datetime
import os
from pathlib import Path
from typing import List

from biliarchiver.cli_tools.utils import read_bvids_from_txt

try:
    from fastapi import FastAPI, HTTPException
except ImportError:
    print("Please install fastapi")
    print("`pip install fastapi`")
    raise


from biliarchiver.cli_tools.get_command import by_favlist, by_series, by_up_videos
from biliarchiver.rest_api.bilivid import BiliVideo, VideoStatus
from biliarchiver.version import BILI_ARCHIVER_VERSION


@asynccontextmanager
async def lifespan(app: FastAPI):
    global pending_queue, other_queue
    pending_queue = BiliVideoQueue()
    other_queue = BiliVideoQueue(maxsize=250)
    print("Loading queue...")
    load_queue()
    print("Queue loaded")
    _video_scheduler = asyncio.create_task(video_scheduler())
    yield
    print("Shutting down...")
    save_queue()
    print("Queue saved")


class BiliVideoQueue(asyncio.Queue):
    def get_all(self) -> List[BiliVideo]:
        return list(self._queue) # type: ignore
    async def get(self) -> BiliVideo:
        return await super().get()
    async def remove(self, video: BiliVideo):
        try:
            self._queue.remove(video) # type: ignore
            return True
        except ValueError:
            return False
    def get_nowait(self) -> BiliVideo:
        return super().get_nowait()
    async def change_status(self, ori_video: BiliVideo, status: VideoStatus):
        await self.remove(ori_video)
        ori_video.status = status
        await self.put(ori_video)

pending_queue: BiliVideoQueue = None # type: ignore
other_queue: BiliVideoQueue = None # type: ignore

app = FastAPI(lifespan=lifespan)

def get_all_items() -> List[BiliVideo]:
    l = pending_queue.get_all() + other_queue.get_all()
    l.sort(key=lambda x: x.added_time)
    return l

@app.get("/")
async def root():
    return {
        "status": "ok",
        "biliarchiver": {"version": BILI_ARCHIVER_VERSION},
        "api": {"version": 1},
        "timestamp": int(datetime.now().timestamp()),
    }


@app.put("/archive/{vid}")
@app.post("/archive/{vid}")
async def add(vid: str):
    video = BiliVideo(vid, status=VideoStatus.pending)
    await pending_queue.put(video)
    return {"success": True, "vid": vid}


@app.get("/archive")
async def get_all():
    all_item = get_all_items()
    return {"success": True, "total_tasks_queued": pending_queue.qsize(), "items": all_item}


@app.get("/archive/{vid}")
async def get_one(vid: str):
    all_item = get_all_items()
    if vid in [v.bvid for v in all_item]:
        v: BiliVideo = [v for v in all_item if v.bvid == vid][0]
        return {"success": True, "vid": vid, "status": v.status, "queue_index": all_item.index(v)}
    return {"success": False, "vid": vid, "status": "not_found"} # TODO


@app.delete("/archive/{vid}")
async def delete(vid: str):
    queue_list = pending_queue.get_all()
    if vid in [v.bvid for v in queue_list]:
        v: BiliVideo = [v for v in queue_list if v.bvid == vid][0]
        if await pending_queue.remove(v):
            return {"success": True, "vid": vid, "result": "removed", "queue_index": queue_list.index(v)}

    return {"success": False, "vid": vid, "status": "not_found"}

async def source_action(fun, source_id: str, TRUNCATE=20):
    try:
        txt_path = await fun(source_id, truncate=TRUNCATE)
    except Exception as e:
        print(f"Failed to call {fun}: {e}")
        raise HTTPException(status_code=500, detail=f"Failed to call {fun}")
    if not isinstance(txt_path, Path):
        raise HTTPException(status_code=500, detail="Failed to get path")

    bvids = read_bvids_from_txt(txt_path)
    txt_path.unlink(missing_ok=True)

    return {"success": True, "bvids": bvids}

@app.post("/get_bvids_by/{source_type}/{source_id}")
async def perform_source_action_from_req(source_type: str, source_id: str):
    # make sure source_id is valid integer
    if not source_id.isdecimal():
        raise HTTPException(status_code=400, detail="Invalid source_id")

    source_id = str(int(source_id))

    fun_mapping = {
        "up_videos": by_up_videos,
        "favlist": by_favlist,
        "series": by_series,
    }

    if source_type not in fun_mapping:
        raise HTTPException(status_code=400, detail="Invalid source_type")

    fun = fun_mapping[source_type]

    assert callable(fun)

    return await source_action(fun, source_id, TRUNCATE=300)

async def video_scheduler():
    while True:
        print("Getting a video URL... If no video URL is printed, the queue is empty.")
        video = await pending_queue.get()
        print(f"Start downloading {video}")

        video.status = VideoStatus.downloading
        await other_queue.put(video)
        downloaded = False
        for _ in range(2):
            try:
                if retcode := await video.down():
                    raise Exception(f"Download failed with retcode {retcode}")
                downloaded = True
                break
            except Exception as e:
                print(e)
                print(f"(down) Retrying {video}...")
                await asyncio.sleep(5)
        if not downloaded:
            await other_queue.change_status(video, VideoStatus.failed)
            print(f"Failed to download {video}")
            continue

        print(f"Start uploading {video}")
        await other_queue.change_status(video, VideoStatus.uploading)
        uploaded = False
        for _ in range(3):
            try:
                retcode = await video.up()
                uploaded = True
                if retcode != 0:
                    raise Exception(f"Upload failed with retcode {retcode}")
                break
            except Exception as e:
                print(e)
                print(f"(up) Retrying {video}...")
                await asyncio.sleep(10)
        if not uploaded:
            await other_queue.change_status(video, VideoStatus.failed)
            print(f"Failed to upload {video}")
            continue

        await other_queue.change_status(video, VideoStatus.finished)
        print(f"Finished {video}")

def save_queue():
    with open("queue.txt", "w") as f:
        while not pending_queue.empty():
            video = pending_queue.get_nowait()
            f.write(str(video) + "\n")


def load_queue():
    try:
        with open("queue.txt", "r") as f:
            while line := f.readline().strip():
                pending_queue.put_nowait(BiliVideo(*line.split("\t")))
    except FileNotFoundError:
        pass



import asyncio
from enum import Enum
import time
from typing import Optional

class VideoStatus(str, Enum):
    pending = "pending"
    downloading = "downloading"
    uploading = "uploading"
    finished = "finished"
    failed = "failed"


class BiliVideo:
    def __init__(self, bvid: str, status: VideoStatus):
        if not bvid.startswith("BV"):
            bvid = "BV" + bvid
        self.added_time = int(time.time())
        self.bvid = bvid
        self.status = status


    def __str__(self) -> str:
        return "\t".join([self.bvid, self.status])

    async def down(self):
        from asyncio import subprocess
        from shlex import quote

        cmd = ["biliarchiver", "down" ,"-i", quote(self.bvid), "-s", "--disable-version-check"]

        process: Optional[subprocess.Process] = None
        try:
            process = await asyncio.create_subprocess_exec(*cmd)
            retcode = await process.wait()
            process = None
        except (KeyboardInterrupt, SystemExit, Exception) as e:
            if process:
                process.terminate()
                await process.wait()
                print("download terminated:", e)
            return -1
        else:
            return retcode
        finally:
            if process:
                process.terminate()
                await process.wait()
                print("download terminated: (finally)")

    async def up(self) -> int:
        from asyncio import subprocess
        from shlex import quote

        cmd = ["biliarchiver", "up" ,"-i", quote(self.bvid), "-d"]

        process: Optional[subprocess.Process] = None
        try:
            process = await subprocess.create_subprocess_exec(*cmd)
            retcode = await process.wait()
            process = None
        except (KeyboardInterrupt, SystemExit, Exception) as e:
            if process:
                process.terminate()
                await process.wait()
                print("upload terminated", e)
            return -1
        else:
            return retcode
        finally:
            if process:
                process.terminate()
                await process.wait()
                print("upload terminated: (finally)")



from typing import Union

_xml_ILLEGAL_CHARS = []
_xml_ILLEGAL_CHARS.extend([bytes([i]).decode('ascii') for i in range(int('00', 16), int('08', 16)+1)])
_xml_ILLEGAL_CHARS.extend([bytes([i]).decode('ascii') for i in range(int('0b', 16), int('0c', 16)+1)])
_xml_ILLEGAL_CHARS.extend([bytes([i]).decode('ascii') for i in range(int('0e', 16), int('1f', 16)+1)])
_xml_ILLEGAL_CHARS.extend(['\x7f'])
# NOTE: The following are non-ASCII characters, which will not appear in string obj in Python.
# range(int('80', 16), int('84', 16)+1)])
# range(int('86', 16), int('9f', 16)+1)])
XML_ILLEGAL_CHARS = _xml_ILLEGAL_CHARS


def _legalize_str(s: str, print_info: bool=True):
    for c in XML_ILLEGAL_CHARS:
        hash_before = hash(s)
        s = s.replace(c, '')
        if print_info and hash(s) != hash_before:
            print(f"Removed XML illegal char \\x{ord(c):02x}")
    return s

def _legalize_list(l: list):
    for i, v in enumerate(l):
        if isinstance(v, str):
            l[i] = _legalize_str(v)
        elif isinstance(v, dict):
            l[i] = _legalize_dict(v)
        elif isinstance(v, list):
            l[i] = _legalize_list(v)
        else:
            pass
    return l

def _legalize_dict(d: dict):
    for k, v in d.items():
        if isinstance(v, str):
            d[k] = _legalize_str(v)
        elif isinstance(v, list):
            d[k] = _legalize_list(v)
        elif isinstance(v, dict):
            d[k] = _legalize_dict(v)
        else:
            pass
    return d

def xml_chars_legalize(obj: Union[dict, str, list]) -> Union[dict, str, list]:
    """ Remove XML illegal characters from a dict, list or str. """
    if isinstance(obj, str):
        return _legalize_str(obj)
    elif isinstance(obj, dict):
        return _legalize_dict(obj)
    elif isinstance(obj, list):
        return _legalize_list(obj)
    else:
        raise TypeError(f"Unexpected type: {type(obj)}")

def _test_xml_chars_legalize():
    _str = '\x0bA\x0cB\vC'
    _str_after = 'ABC'
    assert _legalize_str(_str) == _str_after
    _list = ['\x0bA', '\x0cB', 'C\v']
    _list_after = ['A', 'B', 'C']
    assert _legalize_list(_list) == _list_after
    _dict = {'A': '\x0bA', 'B': '\x0cB', 'C': 'C\v', 'list': _list}
    _dict_after = {'A': 'A', 'B': 'B', 'C': 'C', 'list': _list_after}
    assert _legalize_dict(_dict) == _dict_after
    _dict_in_list_in_dict_in_dict = {"dict": [{'A': '\x0bA', 'B': '\x0cB', 'C': 'C\v', 'list': _list, 'dict': _dict}]}
    _dict_balabala_after = {'dict': [{'A': 'A', 'B': 'B', 'C': 'C', 'list': _list_after, 'dict': _dict_after}]}
    assert _legalize_dict(_dict_in_list_in_dict_in_dict) == _dict_balabala_after
    assert _legalize_str("\x00") == ""

if __name__ == '__main__':
    # TODO: use pytest
    _test_xml_chars_legalize()


import click
from rich.console import Console
from biliarchiver.i18n import _


@click.command(help=click.style(_("从哔哩哔哩下载"), fg="cyan"))
@click.option(
    "--bvids",
    "-i",
    type=click.STRING,
    required=True,
    help=_("空白字符分隔的 bvids 列表（记得加引号），或文件路径"),
)
@click.option(
    "--skip-ia-check",
    "-s",
    is_flag=True,
    default=False,
    show_default=True,
    help=_("不检查 IA 上是否已存在对应 BVID 的 item ，直接开始下载"),
)
@click.option(
    "--from-browser",
    "--fb",
    type=str,
    default=None,
    help=_("从指定浏览器导入 cookies (否则导入 config.json 中的 cookies_file)"),
)
@click.option(
    "--min-free-space-gb",
    type=int,
    default=10,
    help=_("最小剩余空间 (GB)，用超退出"),
    show_default=True,
)
@click.option(
    "--skip-to", type=int, default=0, show_default=True, help=_("跳过文件开头 bvid 的个数")
)
@click.option(
    "--disable-version-check",
    type=bool,
    is_flag=True,
    default=False,
    help=_("禁用 biliarchiver 的 pypi 版本检查")
)
def down(**kwargs):
    from biliarchiver.cli_tools.bili_archive_bvids import _down

    try:
        import asyncio

        asyncio.run(_down(**kwargs))
    except KeyboardInterrupt:
        print("KeyboardInterrupt")
    finally:
        # 显示终端光标
        console = Console()
        console.show_cursor()



from dataclasses import dataclass
import os
import json
from pathlib import Path
from biliarchiver.i18n import _
from biliarchiver.exception import DirNotInitializedError

CONFIG_FILE = "config.json"
BILIBILI_IDENTIFIER_PERFIX = "BiliBili"  # IA identifier 前缀。


class singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


@dataclass
class _Config(metaclass=singleton):
    video_concurrency: int = 3
    part_concurrency: int = 10
    stream_retry: int = 20
    storage_home_dir: Path = Path("bilibili_archive_dir/").expanduser()
    ia_key_file: Path = Path("~/.bili_ia_keys.txt").expanduser()
    cookies_file: Path = Path("~/.cookies.txt").expanduser()

    def __init__(self):
        self.is_right_pwd()
        if not os.path.exists(CONFIG_FILE):
            print(_("{} 不存在，创建中...").format(CONFIG_FILE))
            self.save()
        with open(CONFIG_FILE, "r", encoding="utf-8") as f:
            print(f"Loading {CONFIG_FILE} ...")
            config_file = json.load(f)

        self.video_concurrency: int = config_file["video_concurrency"]
        self.part_concurrency: int = config_file["part_concurrency"]
        self.stream_retry: int = config_file["stream_retry"]

        self.storage_home_dir: Path = Path(config_file["storage_home_dir"]).expanduser()
        self.ia_key_file: Path = Path(config_file["ia_key_file"]).expanduser()
        self.cookies_file: Path = Path(config_file["cookies_file"]).expanduser()

    def save(self):
        with open(CONFIG_FILE, "w", encoding="utf-8") as f:
            json.dump(
                {
                    "video_concurrency": self.video_concurrency,
                    "part_concurrency": self.part_concurrency,
                    "stream_retry": self.stream_retry,
                    "storage_home_dir": str(self.storage_home_dir),
                    "ia_key_file": str(self.ia_key_file),
                    "cookies_file": str(self.cookies_file),
                },
                f,
                ensure_ascii=False,
                indent=4,
            )

    def is_right_pwd(self):
        if not os.path.exists("biliarchiver.home"):
            raise DirNotInitializedError()


config = _Config()



import click
from dataclasses import dataclass
from biliarchiver.i18n import _

@click.command(help=click.style(_("将传入参数写入配置文件"), fg="cyan"))
@click.option("--video_concurrency", "-v", type=click.INT, default=None, help=_("视频下载并发数"))
@click.option("--part_concurrency", "-p", type=click.INT, default=None, help=_("分P下载并发数"))
@click.option("--stream_retry", "-r", type=click.INT, default=None, help=_("流下载重试次数"))
@click.option("--storage_home_dir", "-s", type=click.STRING, default=None, help=_("存储目录"))
@click.option("--ia_key_file", "-i", type=click.STRING, default=None, help=_("IA key文件"))
@click.option("--cookies_file", "-c", type=click.STRING, default=None, help=_("cookies文件"))

def config(**kwargs):
    from biliarchiver.config import _Config
    config = _Config()
    for k, v in kwargs.items():
        if v is not None:
            setattr(config, k, v)
    config.save()
    print(_("配置文件写入成功"))


import asyncio
import os
from pathlib import Path
import re
import traceback

import aiofiles
import httpx
from bilix.download.utils import raise_api_error, req_retry
from bilix.exception import APIError

from bilix.sites.bilibili import api

from rich import print
import json

from bilix.sites.bilibili.downloader import DownloaderBilibili
from bilix.exception import APIResourceError
from biliarchiver.config import BILIBILI_IDENTIFIER_PERFIX
from biliarchiver.config import config
from biliarchiver.utils.identifier import human_readable_upper_part_map
from biliarchiver.i18n import _

# moneky patch
@staticmethod
def _dm2ass_factory(width: int, height: int):
    import functools
    from danmakuC.bilibili import proto2ass
    async def dm2ass(protobuf_bytes: bytes) -> bytes:
        loop = asyncio.get_event_loop()
        f = functools.partial(proto2ass, protobuf_bytes, width, height, font_size=width / 40, ) # type: ignore
        print("using none")
        content = await loop.run_in_executor(None, f) # use default executor (None) instead of bilix._process.SingletonPPE
        return content.encode('utf-8')

    return dm2ass
DownloaderBilibili._dm2ass_factory = _dm2ass_factory


@raise_api_error
async def new_get_subtitle_info(client: httpx.AsyncClient, bvid, cid):
    params = {"bvid": bvid, "cid": cid}
    res = await req_retry(client, "https://api.bilibili.com/x/player/v2", params=params)
    info = json.loads(res.text)
    if info["code"] == -400:
        raise APIError(_("未找到字幕信息"), params)

    # 这里 monkey patch 一下把返回 lan_doc 改成返回 lan，这样生成的字幕文件名就是 语言代码 而不是 中文名 了
    # 例如
    # lan_doc: 中文（中国）
    # lang: zh-CN

    # return [[f'http:{i["subtitle_url"]}', i['lan_doc']] for i in info['data']['subtitle']['subtitles']]
    return [
        [f'http:{i["subtitle_url"]}', i["lan"]]
        for i in info["data"]["subtitle"]["subtitles"]
    ]


api.get_subtitle_info = new_get_subtitle_info


@raise_api_error
async def new_get_video_info(client: httpx.AsyncClient, url: str):
    """
    monkey patch 一下，只使用 API 获取 video_info
    理由：
        - API 目前只支持一般的 AV/BV 视频，可以预防我们不小心下到了番剧/影视剧之类的版权内容
        - 如果缺 url 对应的分P，bilix 那边会报 AssertionError(f"没有找到分P: p{selected_page_num}，请检查输入")
        - 对于一些老视频， _get_video_info_from_html() 经常返回烦人且不稳定的 durl 资源。而 API 会更多请求到 dash 资源（虽然仍有少数视频只有 durl 资源）。
    """
    # print("using api")
    return await api._get_video_info_from_api(client, url)


api.get_video_info = new_get_video_info


async def archive_bvid(
    d: DownloaderBilibili,
    bvid: str,
    *,
    logined: bool = False,
    semaphore: asyncio.Semaphore,
):
    async with semaphore:
        assert d.hierarchy is True, _("hierarchy 必须为 True")  # 为保持后续目录结构、文件命名的一致性
        assert d.client.cookies.get("SESSDATA") is not None, _(
            "sess_data 不能为空"
        )  # 开个大会员呗，能下 4k 呢。
        assert logined is True, _("请先检查 SESSDATA 是否过期，再将 logined 设置为 True")  # 防误操作
        upper_part = human_readable_upper_part_map(string=bvid, backward=True)
        videos_basepath: Path = (
            config.storage_home_dir / "videos" / f"{bvid}-{upper_part}"
        )

        if os.path.exists(videos_basepath / "_all_downloaded.mark"):
            # print(f"{bvid} 所有分p都已下载过了")
            print(_("{} 所有分p都已下载过了").format(bvid))
            return

        url = f"https://www.bilibili.com/video/{bvid}/"
        # 为了获取 pages，先请求一次
        try:
            first_video_info = await api.get_video_info(d.client, url)
        except APIResourceError as e:
            print(_("{} 获取 video_info 失败，原因：{}").format(bvid, e))
            return

        os.makedirs(videos_basepath, exist_ok=True)

        pid = 0
        for page in first_video_info.pages:
            pid += 1  # pid 从 1 开始
            if not page.p_url.endswith(f"?p={pid}"):
                raise NotImplementedError(
                    _("{} 的 P{} 不存在 (可能视频被 UP 主 / B 站删了)，请报告此问题，我们需要这个样本！").format(
                        bvid, pid
                    )
                )

            file_basename = f"{bvid}_p{pid}"
            video_basepath = (
                videos_basepath / f"{BILIBILI_IDENTIFIER_PERFIX}-{file_basename}"
            )
            video_extrapath = video_basepath / "extra"
            if os.path.exists(f"{video_basepath}/_downloaded.mark"):
                print(_("{}: 已经下载过了").format(file_basename))
                continue

            def delete_cache(reason: str = ""):
                if not os.path.exists(video_basepath):
                    return
                _files_in_video_basepath = os.listdir(video_basepath)
                for _file in _files_in_video_basepath:
                    if _file.startswith(file_basename):
                        print(_("{}: {}，删除缓存: {}").format(file_basename, reason, _file))
                        os.remove(video_basepath / _file)

            delete_cache(_("为防出错，清空上次未完成的下载缓存"))
            video_info = await api.get_video_info(d.client, page.p_url)
            print(f"{file_basename}: {video_info.title}...")
            os.makedirs(video_basepath, exist_ok=True)
            os.makedirs(video_extrapath, exist_ok=True)

            old_p_name = video_info.pages[video_info.p].p_name
            old_title = video_info.title

            # 在 d.hierarchy is True 且 title 超长的情况下， bilix 会将 p_name 作为文件名
            video_info.pages[
                video_info.p
            ].p_name = file_basename  # 所以这里覆盖 p_name 为 file_basename
            video_info.title = "iiiiii" * 50  # 然后假装超长标题
            # 这样 bilix 保存的文件名就是我们想要的了（谁叫 bilix 不支持自定义文件名呢）
            # NOTE: p_name 似乎也不宜过长，否则还是会被 bilix 截断。
            # 但是我们以 {bvid}_p{pid} 作为文件名，这个长度是没问题的。

            codec = None
            quality = None
            if video_info.dash:
                # 选择编码 dvh->hev->avc
                # 不选 av0 ，毕竟目前没几个设备能拖得动
                codec_candidates = ["dvh", "hev", "avc"]
                for codec_candidate in codec_candidates:
                    for media in video_info.dash.videos:
                        if media.codec.startswith(codec_candidate):
                            codec = media.codec
                            quality = media.quality
                            print(f'{file_basename}: "{codec}" "{media.quality}" ...')
                            break
                    if codec is not None:
                        break
                assert (
                    codec is not None and quality is not None
                ), f"{file_basename}: " + _("没有 dvh、avc 或 hevc 编码的视频")
            elif video_info.other:
                # print(f"{file_basename}: 未解析到 dash 资源，交给 bilix 处理 ...")
                print("{file_basename}: " + _("未解析到 dash 资源，交给 bilix 处理 ..."))
                codec = ""
                quality = 0
            else:
                raise APIError("{file_basename}: " + _("未解析到视频资源"), page.p_url)

            assert codec is not None
            assert isinstance(quality, (int, str))

            cor1 = d.get_video(
                page.p_url,
                video_info=video_info,
                path=video_basepath,
                quality=quality,  # 选择最高画质
                codec=codec,  # 编码
                # 下载 ass 弹幕(bilix 会自动调用 danmukuC 将 pb 弹幕转为 ass)、封面、字幕
                # 弹幕、封面、字幕都会被放进 extra 子目录里，所以需要 d.hierarchy is True
                dm=True,
                image=True,
                subtitle=True,
            )
            # 下载原始的 pb 弹幕
            cor2 = d.get_dm(page.p_url, video_info=video_info, path=video_extrapath)
            # 下载视频超详细信息（BV 级别，不是分 P 级别）
            cor3 = download_bilibili_video_detail(
                d.client, bvid, f"{video_extrapath}/{file_basename}.info.json"
            )
            # 下载视频评论。有些视频关闭了评论会获取不到。
            cor4 = download_bilibili_video_replies(
                d.client, video_info.bvid, video_info.aid,
                f"{video_extrapath}/{file_basename}.replies.json"
            )
            coroutines = [cor1, cor2, cor3, cor4]
            tasks = [asyncio.create_task(cor) for cor in coroutines]
            results = await asyncio.gather(*tasks, return_exceptions=True)
            for result, cor in zip(results, coroutines):
                if isinstance(result, Exception):
                    print(_("出错，其他任务完成后将抛出异常..."))
                    for task in tasks:
                        task.cancel()
                    await asyncio.sleep(3)
                    traceback.print_exception(result)
                    raise result

            if codec.startswith("hev") and not os.path.exists(
                video_basepath / f"{file_basename}.mp4"
            ):
                # 如果有下载缓存文件（以 file_basename 开头的文件），说明这个 hevc 的 dash 资源存在，只是可能因为网络之类的原因下载中途失败了
                delete_cache(_("下载出错"))

                # 下载缓存文件都不存在，应该是对应的 dash 资源根本就没有，一些老视频会出现这种情况。
                # 换 avc 编码
                print(
                    _("{}: 视频文件没有被下载？也许是 hevc 对应的 dash 资源不存在，尝试 avc ……").format(
                        file_basename
                    )
                )
                assert video_info.dash is not None
                for media in video_info.dash.videos:
                    if media.codec.startswith("avc"):
                        codec = media.codec
                        print(f'{file_basename}: "{codec}" "{media.quality}" ...')
                        break
                cor4 = d.get_video(
                    page.p_url,
                    video_info=video_info,
                    path=video_basepath,
                    quality=0,  # 选择最高画质
                    codec=codec,  # 编码
                    # 下载 ass 弹幕(bilix 会自动调用 danmukuC 将 pb 弹幕转为 ass)、封面、字幕
                    # 弹幕、封面、字幕都会被放进 extra 子目录里，所以需要 d.hierarchy is True
                    dm=True,
                    image=True,
                    subtitle=True,
                )
                await cor4

            assert os.path.exists(
                video_basepath / f"{file_basename}.mp4"
            ) or os.path.exists(video_basepath / f"{file_basename}.flv")

            # 还原为了自定义文件名而做的覆盖
            video_info.pages[video_info.p].p_name = old_p_name
            video_info.title = old_title

            # 单 p 下好了
            async with aiofiles.open(
                f"{video_basepath}/_downloaded.mark", "w", encoding="utf-8"
            ) as f:
                await f.write("")

        # bv 对应的全部 p 下好了
        async with aiofiles.open(
            f"{videos_basepath}/_all_downloaded.mark", "w", encoding="utf-8"
        ) as f:
            await f.write("")


async def download_bilibili_video_detail(client, bvid, filepath):
    if os.path.exists(filepath):
        print(_("{} 的视频详情已存在").format(bvid))
        return
    # url = 'https://api.bilibili.com/x/web-interface/view'
    url = "https://api.bilibili.com/x/web-interface/view/detail"  # 超详细 API（BV 级别，不是分 P 级别）
    params = {"bvid": bvid}
    r = await req_retry(client, url, params=params, follow_redirects=True)
    r.raise_for_status()
    r_json = r.json()
    assert r_json["code"] == 0, _("{} 的视频详情获取失败").format(bvid)

    async with aiofiles.open(filepath, "w", encoding="utf-8") as f:
        # f.write(json.dumps(r.json(), indent=4, ensure_ascii=False))
        await f.write(r.text)
    print(_("{} 的视频详情已保存").format(bvid))


async def download_bilibili_video_replies(client, bvid, aid, filepath):
    for i in range(2):
        try:
            await _download_bilibili_video_replies(client, bvid, aid, filepath)
            return True
        except Exception as e:
            print(e, "retrying...")
    
    print(_("{} 的视频回复获取失败").format(bvid))

async def _download_bilibili_video_replies(client, bvid, aid, filepath):
    """ 仅下载第一页，20 个热评 """
    if os.path.exists(filepath):
        print(_("{} 的视频回复已存在").format(bvid))
        return

    url = "https://api.bilibili.com/x/v2/reply"
    params = {
        "type": 1,
        "oid": aid,
        "sort": 1, # 0：按时间, 1：按点赞数, 2：按回复数
        "ps": 20,
        "pn": 1
    }
    r = await req_retry(client, url, params=params, follow_redirects=True)
    r.raise_for_status()
    r_json = r.json()
    assert r_json["code"] == 0, _("{} 的视频回复获取失败").format(bvid)

    async with aiofiles.open(filepath, "w", encoding="utf-8") as f:
        # f.write(json.dumps(r.json(), indent=4, ensure_ascii=False))
        await f.write(r.text)
    print(_("{} 的视频评论已保存").format(bvid))




import ctypes
import os
import platform
from typing import Union
from pathlib import Path


def get_free_space(path: Union[Path,str]) -> int:
    """Return folder/drive free space (in bytes)."""
    if not isinstance(path, Path):
        path = Path(path)
    if not os.path.exists(path):
        raise FileNotFoundError(f'{path} File Not Found')
    if platform.system() == 'Windows':
        free_bytes = ctypes.c_ulonglong(0)
        ctypes.windll.kernel32.GetDiskFreeSpaceExW(ctypes.c_wchar_p(os.path.abspath(path)), None, None, ctypes.pointer(free_bytes)) # type: ignore
        return free_bytes.value
    else:
        st = os.statvfs(path)
        return st.f_bavail * st.f_frsize



import asyncio
import json
import os
import re
import time
from pathlib import Path

import click
import requests
from bilix.sites.bilibili import api
from click_option_group import optgroup
from httpx import AsyncClient
from rich import print

from biliarchiver.i18n import _, ngettext


async def by_series(url_or_sid: str, truncate: int = int(1e10)) -> Path:
    """
    truncate: do noting
    """
    sid = sid = (
        re.search(r"sid=(\d+)", url_or_sid).groups()[0]
        if url_or_sid.startswith("http")
        else url_or_sid
    )  # type: ignore
    client = AsyncClient(**api.dft_client_settings)
    print(_("正在获取 {sid} 的视频列表……").format(sid=sid))
    col_name, up_name, bvids = await api.get_collect_info(client, sid)
    filepath = f"bvids/by-sapce_fav_season/sid-{sid}-{int(time.time())}.txt"
    os.makedirs(os.path.dirname(filepath), exist_ok=True)
    abs_filepath = os.path.abspath(filepath)
    with open(abs_filepath, "w", encoding="utf-8") as f:
        for bv_id in bvids:
            f.write(f"{bv_id}" + "\n")
    # print(f"已获取 {col_name}（{up_name}）的 {len(bvids)} 个视频")
    count = len(bvids)
    print(
        ngettext(
            "已获取 {}（{}）的一个视频",
            "已获取 {}（{}）的 {count} 个视频",
            count,
        ).format(col_name, up_name, count=count)
    )
    print(_("存储到 {}").format(abs_filepath))
    return Path(abs_filepath)


def by_ranking(rid: int) -> Path:
    bilibili_ranking_api = "https://api.bilibili.com/x/web-interface/ranking/v2"
    bilibili_ranking_params = {"rid": rid, "type": "all"}

    r = requests.get(bilibili_ranking_api, params=bilibili_ranking_params)
    r.raise_for_status()
    ranking_json = json.loads(r.text)
    assert ranking_json["code"] == 0  # 0 为成功（HTTP 200 不能信）

    ranking = ranking_json["data"]["list"]
    bvids = []
    for video_info in ranking:
        # print(video_info['title'], video_info['bvid'], video_info['pic'])
        bvid = video_info["bvid"]
        bvids.append(bvid)

    import datetime

    datetime.date.today()
    os.makedirs("bvids", exist_ok=True)

    bvids_filepath = f"bvids/by-ranking/rid-{rid}/rid-{rid}-{int(time.time())}.txt"
    os.makedirs(os.path.dirname(bvids_filepath), exist_ok=True)
    with open(bvids_filepath, "w", encoding="utf-8") as f:
        for bvid in bvids:
            f.write(f"{bvid}" + "\n")
    abs_filepath = os.path.abspath(bvids_filepath)
    print(
        ngettext("已保存一个 bvid 到 {}", "已保存 {count} 个 bvid 到 {}", len(bvids)).format(
            abs_filepath, count=len(bvids)
        )
    )
    return Path(abs_filepath)


async def by_up_videos(url_or_mid: str, truncate: int = int(1e10)) -> Path:
    """频率高了会封"""

    if isinstance(url_or_mid, int):
        mid = str(url_or_mid)
    elif url_or_mid.startswith("http"):
        mid = re.findall(r"/(\d+)", url_or_mid)[0]
    else:
        mid = url_or_mid

    assert isinstance(mid, str)
    assert mid.isdigit(), _("mid 应是数字字符串")

    client = AsyncClient(**api.dft_client_settings)

    from biliarchiver.config import config

    if config.cookies_file.exists():
        from biliarchiver.cli_tools.bili_archive_bvids import update_cookies_from_file
        update_cookies_from_file(client, config.cookies_file)
        delay = 3
    else:
        print(_("cookies 文件不存在: {}").format(config.cookies_file))
        delay = 10

    ps = 30  # 每页视频数，最小 1，最大 50，默认 30
    order = "pubdate"  # 默认为pubdate 最新发布：pubdate 最多播放：click 最多收藏：stow
    keyword = ""  # 搜索关键词
    bv_ids = []
    pn = 1
    print(ngettext("获取第 {} 页...", "获取第 {} 页...", pn).format(pn))
    up_name, total_size, bv_ids_page = await api.get_up_video_info(
        client, mid, pn, ps, order, keyword
    )
    bv_ids += bv_ids_page
    # print(f"{mid} {up_name} 共 {total_size} 个视频. (如果最新的视频为合作视频的非主作者，UP 名可能会识别错误，但不影响获取 bvid 列表)")
    print(
        ngettext("{} {} 共 {} 个视频.", "{} {} 共 {} 个视频.", total_size).format(
            mid, up_name, total_size
        )
    )
    print(_("（如果最新的视频为合作视频的非主作者，UP 名可能会识别错误，但不影响获取 bvid 列表)"))
    while pn < total_size / ps:
        pn += 1
        print(ngettext("获取第 {} 页", "获取第 {} 页", pn).format(pn))
        await asyncio.sleep(delay)
        _x, _y, bv_ids_page = await api.get_up_video_info(client, mid, pn, ps, order, keyword)
        bv_ids += bv_ids_page

        if len(bv_ids) >= truncate:
            print("truncate at", truncate)
            break

    print(mid, up_name, total_size)
    await client.aclose()
    assert len(bv_ids) == len(set(bv_ids)), _("有重复的 bv_id")
    assert total_size == len(bv_ids), _("视频总数不匹配")
    filepath = f"bvids/by-up_videos/mid-{mid}-{int(time.time())}.txt"
    os.makedirs(os.path.dirname(filepath), exist_ok=True)
    abs_filepath = os.path.abspath(filepath)
    with open(abs_filepath, "w", encoding="utf-8") as f:
        for bv_id in bv_ids:
            f.write(f"{bv_id}" + "\n")
    print(
        ngettext("已保存一个 bvid 到 {}", "已保存 {count} 个 bvid 到 {}", len(bv_ids)).format(
            abs_filepath, count=len(bv_ids)
        )
    )
    return Path(abs_filepath)


def by_popular_precious():
    API_URL = "https://api.bilibili.com/x/web-interface/popular/precious"
    r = requests.get(API_URL)
    r.raise_for_status()
    popular_precious_json = json.loads(r.text)
    assert popular_precious_json["code"] == 0
    bvids = []
    for video_info in popular_precious_json["data"]["list"]:
        bvid = video_info["bvid"]
        bvids.append(bvid)
    filepath = f"bvids/by-popular_precious/{int(time.time())}.txt"
    os.makedirs(os.path.dirname(filepath), exist_ok=True)
    abs_filepath = os.path.abspath(filepath)
    with open(abs_filepath, "w", encoding="utf-8") as f:
        f.write("\n".join(bvids))
    # print(f"已保存 {len(bvids)} 个 bvid 到 {abs_filepath}")
    print(
        ngettext("已保存一个 bvid 到 {}", "已保存 {count} 个 bvid 到 {}", len(bvids)).format(
            abs_filepath, count=len(bvids)
        )
    )


def by_popular_series_one(number: int):
    API_URL = "https://api.bilibili.com/x/web-interface/popular/series/one"
    params = {"number": number}
    r = requests.get(API_URL, params=params)
    r.raise_for_status()
    popular_series_json = json.loads(r.text)
    assert popular_series_json["code"] == 0
    bvids = []
    for video_info in popular_series_json["data"]["list"]:
        bvid = video_info["bvid"]
        bvids.append(bvid)
    filepath = f"bvids/by-popular_series/s{number}-{int(time.time())}.txt"
    os.makedirs(os.path.dirname(filepath), exist_ok=True)
    abs_filepath = os.path.abspath(filepath)
    with open(abs_filepath, "w", encoding="utf-8") as f:
        f.write("\n".join(bvids))
    # print(f"已保存 {len(bvids)} 个 bvid 到 {abs_filepath}")
    print(
        ngettext("已保存一个 bvid 到 {}", "已保存 {count} 个 bvid 到 {}", len(bvids)).format(
            abs_filepath, count=len(bvids)
        )
    )


def not_got_popular_series() -> list[int]:
    API_URL = "http://api.bilibili.com/x/web-interface/popular/series/list"
    got_series = []
    os.makedirs("bvids/by-popular_series", exist_ok=True)
    for filename in os.listdir("bvids/by-popular_series"):
        if filename.endswith(".txt"):
            # s{number}-{int(time.time())}.txt
            got_series.append(int(filename.split("-")[0][1:]))
    r = requests.get(API_URL)
    r.raise_for_status()
    popular_series_json = json.loads(r.text)
    assert popular_series_json["code"] == 0
    max_series_number = popular_series_json["data"]["list"][0]["number"]
    series_not_got = []
    for i in range(1, max_series_number + 1):
        if i not in got_series:
            series_not_got.append(i)
    return series_not_got


async def by_favlist(url_or_fid: str, truncate: int = int(1e10)) -> Path:
    if url_or_fid.startswith("http"):
        fid = re.findall(r"fid=(\d+)", url_or_fid)[0]
    else:
        fid = url_or_fid

    client = AsyncClient(**api.dft_client_settings)
    PAGE_SIZE = 20
    media_left = None
    total_size = None
    bvids = []
    page_num = 1
    while media_left is None or media_left > 0:
        # bilix 的收藏夹获取有 bug
        fav_name, up_name, total_size, available_bvids = await api.get_favour_page_info(
            client=client, url_or_fid=fid, pn=page_num, ps=PAGE_SIZE, keyword=""
        )
        bvids.extend(available_bvids)
        if media_left is None:
            print(f"fav_name: {fav_name}, up_name: {up_name}, total_size: {total_size}")
        media_left = total_size - PAGE_SIZE * page_num
        print(
            ngettext("还剩 ~{} 页", "还剩 ~{} 页", media_left // PAGE_SIZE).format(
                media_left // PAGE_SIZE
            ),
            end="\r",
        )

        if len(bvids) >= truncate:
            print("truncate at", truncate)
            break

        await asyncio.sleep(2)
        page_num += 1
    await client.aclose()
    assert total_size is not None
    assert len(bvids) == len(set(bvids)), _("有重复的 bvid")
    print(
        ngettext(
            "{} 个有效视频，{} 个失效视频",
            "{} 个有效视频，{} 个失效视频",
            total_size,
        ).format(len(bvids), total_size - len(bvids))
    )
    filepath = f"bvids/by-favour/fid-{fid}-{int(time.time())}.txt"
    os.makedirs(os.path.dirname(filepath), exist_ok=True)
    abs_filepath = os.path.abspath(filepath)
    with open(abs_filepath, "w", encoding="utf-8") as f:
        f.write("\n".join(bvids))
        f.write("\n")
    print(
        ngettext("已保存一个 bvid 到 {}", "已保存 {count} 个 bvid 到 {}", len(bvids)).format(
            abs_filepath, count=len(bvids)
        )
    )

    return Path(abs_filepath)


async def main(
    series: str,
    ranking: str,
    rid: str,
    up_videos: str,
    popular_precious: bool,
    popular_series: bool,
    popular_series_number: int,
    all_popular_series: bool,
    favlist: str,
):
    if ranking:
        by_ranking(rid)
    if up_videos:
        await by_up_videos(up_videos)
    if popular_precious:
        by_popular_precious()
    if popular_series:
        if all_popular_series:
            for number in not_got_popular_series():
                time.sleep(3)
                by_popular_series_one(number)
        else:
            by_popular_series_one(popular_series_number)
    if series:
        await by_series(series)
    if favlist:
        await by_favlist(favlist)


class URLorIntParamType(click.ParamType):
    def __init__(self, name):
        self.name = "URL|" + name

    def convert(self, value, param, ctx):
        # Simple regex to check if value might be a URL
        # (just checking if it starts with http:// or https://)
        url_pattern = re.compile(r"^https?://")

        # If value matches URL pattern or is a digit, return value
        if url_pattern.match(value) or value.isdigit():
            return value

        # If value doesn't match any, raise an error
        self.fail(f"{value!r} is not a valid {self.name}", param, ctx)


@click.command(
    short_help=click.style(_("批量获取 BV 号"), fg="cyan"),
    help=_("请通过 flag 指定至少一种批量获取 BV 号的方式。多个不同组的 flag 同时使用时，将会先后通过不同方式获取。"),
)
@optgroup.group(_("合集"))
@optgroup.option(
    "--series",
    "-s",
    help=click.style(_("合集或视频列表内视频"), fg="red"),
    type=URLorIntParamType("sid"),
)
@optgroup.group(_("排行榜"))
@optgroup.option(
    "--ranking",
    "-r",
    help=click.style(_("排行榜（全站榜，非个性推荐榜）"), fg="yellow"),
    is_flag=True,
)
@optgroup.option(
    "--rid",
    "--ranking-id",
    default=0,
    show_default=True,
    help=click.style(_("目标排行 rid，0 为全站排行榜。rid 等于分区的 tid"), fg="yellow"),
    type=int,
)
@optgroup.group(_("UP 主"))
@optgroup.option(
    "--up-videos",
    "-u",
    help=click.style(_("UP 主用户页投稿"), fg="cyan"),
    type=URLorIntParamType("mid"),
)
@optgroup.group(_("入站必刷"))
@optgroup.option(
    "--popular-precious",
    help=click.style(_("入站必刷，更新频率低"), fg="bright_red"),
    is_flag=True,
)
@optgroup.group(_("每周必看"))
@optgroup.option(
    "--popular-series",
    "-p",
    help=click.style(_("每周必看，每周五晚 18:00 更新"), fg="magenta"),
    is_flag=True,
)
@optgroup.option(
    "--popular-series-number",
    default=1,
    type=int,
    show_default=True,
    help=click.style(_("获取第几期（周）"), fg="magenta"),
)
@optgroup.option(
    "--all-popular-series",
    help=click.style(_("自动获取全部的每周必看（增量）"), fg="magenta"),
    is_flag=True,
)
@optgroup.group(_("收藏夹"))
@optgroup.option(
    "--favlist",
    "--fav",
    help=click.style(_("用户收藏夹"), fg="green"),
    type=URLorIntParamType("fid"),
)
def get(**kwargs):
    if (
        not kwargs["favlist"]
        and not kwargs["series"]
        and not kwargs["ranking"]
        and not kwargs["up_videos"]
        and not kwargs["popular_precious"]
        and not kwargs["popular_series"]
    ):
        click.echo(click.style(_("ERROR: 请指定至少一种获取方式。"), fg="red"))
        click.echo(get.get_help(click.Context(get)))
        return
    asyncio.run(main(**kwargs))
